"""dataframe.py: classes for storing and retrieving dataframes as RESQML objects.

   note that this module uses the obj_Grid2dRepresentation class in a way that was not envisaged
   when the RESQML standard was defined; software that does not use resqpy is unlikely to be
   able to do much with data stored in this way
"""

version = '6th May 2021'

import logging
import warnings

log = logging.getLogger(__name__)
log.debug(f'dataframe.py version {version}')

import numpy as np
import pandas as pd

import resqpy.property as rqp
import resqpy.surface as rqs
import resqpy.time_series as rqts
import resqpy.crs as rqc
import resqpy.olio.xml_et as rqet

# todo: add support for building an ensemble of dataframes using the same mesh support


class DataFrame:
   """Class for storing and retrieving a pandas dataframe of numerical data as a RESQML property.

   notes:
      actual values are stored either as z values in a Mesh (Grid2d) object, or as a property on
      such a mesh when multiple raalizations are in use; a regular Mesh object is created to act
      as a supporting representation; columns are mapped onto I and rows onto J; if a property is
      used then the indexable elements are 'nodes'; column titles are stored in a related StringLookup
      object, indexed by column number; column units are optionally treated in the same way (uom for
      the property is generally set to Euc); all values are stored as floats; use the derived TimeTable
      class if rows relate to steps in a TimeSeries; use the derived RelPerm class if the rows relate to
      relative permeability data
   """

   def __init__(
         self,
         model,
         support_root = None,  # deprecated
         uuid = None,
         df = None,
         uom_list = None,
         realization = None,
         title = 'dataframe',
         column_lookup_uuid = None,
         uom_lookup_uuid = None,
         extra_metadata = None):
      """Create a new Dataframe object from either a previously stored property or a pandas dataframe.

      arguments:
         model (model.Model): the model to which the new Dataframe will be attached
         support_root (lxml.Element, DEPRECATED): use uuid instead
         uuid (uuid.UUID, optional): the uuid of an existing Grid2dRepresentation
            object acting as support for a dataframe property (or holding the dataframe as z values)
         df (pandas.DataFrame, optional): a dataframe from which the new Dataframe is to be created;
            if both uuid (or support_root) and df are supplied, realization must not be None and a new
            realization property will be created
         uom_list (list of str, optional): a list holding the units of measure for each
            column; if present, length of list must match number of columns in df; ignored if
            uuid or support_root is not None
         realization (int, optional): if present, the realization number of the RESQML property
            holding the dataframe
         title (str, default 'dataframe'): used as the citation title for the Mesh (and property);
            ignored if uuid or support_root is not None
         column_lookup_uuid (uuid, optional): if present, the uuid of a string lookup table holding
            the column names; if present, the contents and order of the table must match the columns
            in the dataframe; if absent, a new lookup table will be created; ignored if support_root
            is not None
         uom_lookup_uuid (uuid, optional): if present, the uuid of a string lookup table holding
            the units of measure for each column; if None and uom_list is present, a new table
            will be created; if both uom_list and uom_lookup_uuid are present, their contents
            must match; ignored if support_root is not None
         extra_metadata (dict, optional): if present, a dictionary of extra metadata items, str: str;
            ignored if uuid (or support_root) is not None

      returns:
         a newly created Dataframe object

      notes:
         when initialising from an existing RESQML object, the supporting mesh and its property should
         have been originally created using this class; when working with ensembles, each object of this
         class will only handle the data for one realization, though they may share a common support_root
      """

      assert uuid is not None or support_root is not None or df is not None
      assert (uuid is None and support_root is None) or df is None or realization is not None

      if uuid is None:
         if support_root is not None:
            warnings.warn("support_root parameter is deprecated, use uuid instead", DeprecationWarning)
            uuid = rqet.uuid_for_part_root(support_root)
      else:
         support_root = model.root_for_uuid(uuid)

      self.model = model
      self.df = None
      self.n_rows = self.n_cols = 0
      self.uom_list = None
      self.realization = realization
      self.title = title
      self.mesh = None  # only generated when needed for write_hdf5(), create_xml()
      self.pc = None  # property collection; only generated when needed for write_hdf5(), create_xml()
      self.column_lookup_uuid = column_lookup_uuid
      self.column_lookup = None  # string lookup table mapping column index (0 based) to column name
      self.uom_lookup_uuid = uom_lookup_uuid
      self.uom_lookup = None  # string lookup table mapping column index (0 based) to uom
      self.extra_metadata = extra_metadata

      if uuid is not None:
         assert rqet.node_type(support_root) == 'obj_Grid2dRepresentation'
         self.mesh = rqs.Mesh(self.model, uuid = uuid)
         self.extra_metadata = self.mesh.extra_metadata
         assert 'dataframe' in self.extra_metadata and self.extra_metadata['dataframe'] == 'true'
         self.title = self.mesh.title
         self.n_rows, self.n_cols = self.mesh.nj, self.mesh.ni
         cl_uuid = self.model.uuid(obj_type = 'StringTableLookup', related_uuid = uuid, title = 'dataframe columns')
         assert cl_uuid is not None, 'column name lookup table not found for dataframe'
         self.column_lookup = rqp.StringLookup(self.model, uuid = cl_uuid)
         self.column_lookup_uuid = self.column_lookup.uuid
         assert self.column_lookup.length() == self.n_cols
         ul_uuid = self.model.uuid(obj_type = 'StringTableLookup', related_uuid = uuid, title = 'dataframe units')
         if ul_uuid is not None:
            self.uom_lookup = rqp.StringLookup(self.model, uuid = ul_uuid)
            self.uom_lookup_uuid = self.uom_lookup.uuid
            self.uom_list = self.uom_lookup.get_list()
         da = self.mesh.full_array_ref()[..., 2]  # dataframe data as 2D numpy array, defaulting to z values in mesh
         existing_pc = rqp.PropertyCollection(support = self.mesh)
         existing_count = 0 if existing_pc is None else existing_pc.number_of_parts()
         if df is None:  # existing dara, either in mesh or property
            if existing_count > 0:  # use property data instead of z values
               if existing_count == 1:
                  if self.realization is not None:
                     assert existing_pc.realization_for_part(existing_pc.singleton()) == self.realization
               else:
                  assert self.realization is not None, 'no realization specified when accessing ensemble dataframe'
               da = existing_pc.single_array_ref(realization = self.realization)
               assert da is not None and da.ndim == 2 and da.shape == (self.n_rows, self.n_cols)
            else:
               assert realization is None
            self.df = pd.DataFrame(da, columns = self.column_lookup.get_list())
         else:  # both support_root and df supplied: add a new realisation
            if existing_count > 0:
               assert existing_pc.singleton(
                  realization = self.realization) is None, 'dataframe realization already exists'
            self.df = df.copy()
            assert len(self.df) == self.n_rows
            assert len(self.df.columns) == self.n_rows
      else:
         assert df is not None, 'no dataframe (or support root) provided when instantiating DataFrame object'
         self.df = df.copy()
         # todo: check data type of columns – restrict to numerical data
         self.n_rows = len(self.df)
         self.n_cols = len(self.df.columns)
         if column_lookup_uuid is not None:
            self.column_lookup = rqp.StringLookup(self.model, uuid = column_lookup_uuid)
            assert self.column_lookup is not None
            assert self.column_lookup.length() == self.n_cols
            assert all(self.df.columns == self.column_lookup.get_list())  # exact match of column names required!
         if uom_lookup_uuid is not None:
            self.uom_lookup = rqp.StringLookup(self.model, uuid = uom_lookup_uuid)
            assert self.uom_lookup is not None
         if uom_list is not None:
            assert len(uom_list) == self.n_cols
            self.uom_list = uom_list.copy()
            if self.uom_lookup is not None:
               assert self.uom_list == self.uom_lookup.get_list()
         elif self.uom_lookup is not None:
            self.uom_list = self.uom_lookup.get_list()

   def dataframe(self):
      """Returns the Dataframe as a pandas DataFrame."""

      return self.df

   def column_uom(self, col_index):
      """Returns units of measure for the specified column, or Euc if no units present."""

      if self.units_table is None:
         return 'Euc'
      assert 0 <= col_index < self.n_cols, 'column index out of range'
      return self.units_table.get_string(col_index)

   def write_hdf5_and_create_xml(self):
      """Write dataframe data to hdf5 file and create xml for RESQML objects to represent dataframe."""

      self._set_mesh_from_df()  # writes hdf5 data and creates xml for mesh (and property)

      if self.column_lookup is None:
         self.column_lookup = rqp.StringLookup(self.model,
                                               int_to_str_dict = dict(enumerate(self.df.columns)),
                                               title = 'dataframe columns')
         self.column_lookup_uuid = self.column_lookup.uuid
         sl_node = self.column_lookup.create_xml()
      else:
         sl_node = self.column_lookup.root
      if sl_node is not None:
         self.model.create_reciprocal_relationship(self.mesh.root, 'destinationObject', sl_node, 'sourceObject')

      if self.uom_list and self.uom_lookup is None:
         self.uom_lookup = rqp.StringLookup(self.model,
                                            int_to_str_dict = dict(enumerate(self.uom_list)),
                                            title = 'dataframe units')
         self.uom_lookup_uuid = self.uom_lookup.uuid
         ul_node = self.uom_lookup.create_xml()
      elif self.uom_lookup is not None:
         ul_node = self.uom_lookup.root
      else:
         ul_node = None
      if ul_node is not None:
         self.model.create_reciprocal_relationship(self.mesh.root, 'destinationObject', ul_node, 'sourceObject')

   def _set_mesh_from_df(self):
      """Creates Mesh object; called before writing to hdf5 or creating xml."""
      # note: actual data is stored in related Property if realization number is present, directly in Mesh otherwise

      assert self.n_rows == len(self.df)
      assert self.n_cols == len(self.df.columns)

      if self.mesh is None:
         origin = (0.0, 0.0, 0.0)
         dxyz_dij = np.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0]])
         crs_uuids = self.model.uuids(obj_type = 'LocalDepth3dCrs')
         if len(crs_uuids) == 0:
            crs = rqc.Crs(self.model)
            crs.create_xml()
            crs_uuid = crs.uuid
         else:  # use any available crs
            crs_uuid = crs_uuids[0]
         if self.realization is None:
            self.mesh = rqs.Mesh(self.model,
                                 mesh_flavour = 'reg&z',
                                 ni = self.n_cols,
                                 nj = self.n_rows,
                                 dxyz_dij = dxyz_dij,
                                 origin = origin,
                                 z_values = np.array(self.df),
                                 crs_uuid = crs_uuid)
         else:
            self.mesh = rqs.Mesh(self.model,
                                 mesh_flavour = 'regular',
                                 ni = self.n_cols,
                                 nj = self.n_rows,
                                 dxyz_dij = dxyz_dij,
                                 origin = origin,
                                 crs_uuid = crs_uuid)
         self.mesh.write_hdf5()
         mesh_root = self.mesh.create_xml(title = self.title)
         rqet.create_metadata_xml(mesh_root, {'dataframe': 'true'})
         if self.realization is not None:
            self.pc = rqp.PropertyCollection()
            self.pc.set_support(support = self.mesh)
            dataframe_pk_uuid = self.model.uuid(obj_type = 'PropertyKind', title = 'dataframe')
            if dataframe_pk_uuid is None:
               dataframe_pk = rqp.PropertyKind(self.model, title = 'dataframe', example_uom = 'Euc')
               dataframe_pk.create_xml()
               dataframe_pk_uuid = dataframe_pk.uuid
            self.pc.add_cached_array_to_imported_list(np.array(self.df),
                                                      'dataframe',
                                                      self.title,
                                                      uom = 'Euc',
                                                      property_kind = 'dataframe',
                                                      local_property_kind_uuid = dataframe_pk_uuid,
                                                      realization = self.realization,
                                                      indexable_element = 'nodes')
            self.pc.write_hdf5_for_imported_list()
            self.pc.create_xml_for_imported_list_and_add_parts_to_model()


class TimeTable(DataFrame):
   """Class for storing and retrieving a pandas dataframe where rows relate to steps in a time series.

   note:
      inherits from DataFrame class
   """

   def __init__(
         self,
         model,
         support_root = None,  # deprecated
         uuid = None,
         df = None,
         uom_list = None,
         realization = None,
         time_series = None,
         title = 'timetable',
         column_lookup_uuid = None,
         uom_lookup_uuid = None):
      """Create a new TimeTable object from either a previously stored property or a pandas dataframe.

      arguments:
         time_series (resqpy.time_series.TimeSeries): the time series which rows in the dataframe relate to;
            required if initialising from a dataframe, ignored otherwise

      note:
         see DataFrame class docstring for details of other arguments
      """

      # todo: add option to set up time series from a column in the dataframe?

      assert uuid is not None or support_root is not None or (df is not None and time_series is not None)

      super().__init__(model,
                       uuid = uuid,
                       support_root = support_root,
                       df = df,
                       uom_list = uom_list,
                       realization = realization,
                       title = title,
                       column_lookup_uuid = column_lookup_uuid,
                       uom_lookup_uuid = uom_lookup_uuid)
      if uuid is not None or support_root is not None:
         ts_uuid = self.model.uuid(obj_type = 'TimeSeries', related_uuid = self.mesh.uuid)
         assert ts_uuid is not None, 'no time series related to mesh holding dataframe'
         self.ts = rqts.TimeSeries(self.model, uuid = ts_uuid)
      else:
         assert time_series is not None
         assert time_series.number_of_timestamps() == self.n_rows
         self.ts = time_series

   def time_series(self):
      """Returns the TimeSeries object in use by the time table."""

      return self.ts

   def write_hdf5_and_create_xml(self):
      """Write time table data to hdf5 file and create xml for RESQML objects to represent dataframe."""

      super().write_hdf5_and_create_xml()
      # note: time series xml must be created before calling this method
      self.model.create_reciprocal_relationship(self.mesh.root, 'destinationObject', self.ts.root, 'sourceObject')


def dataframe_parts_in_model(model, timetables = None, title = None, related_uuid = None):
   """Returns list of part names within model that are representing DataFrame support objects.

   arguments:
      model (model.Model): the model to be inspected for dataframes
      timetables (boolean or None): if True, only TimeTable dataframe parts will be included; if False
         only DataFrame parts that are not representing TimeTable objects will be included; if None,
         both parts for both types of dataframe will be included
      title (str, optional): if present, only parts with a citation title exactly matching will be
         included
      related_uuid (uuid, optional): if present, only parts relating to this uuid are included

   returns:
      list of str, each element in the list is a part name, within model, which is representing the
      support for a DataFrame object
   """

   df_parts_list = model.parts(obj_type = 'Grid2dRepresentation',
                               title = title,
                               extra = {'dataframe': 'true'},
                               related_uuid = related_uuid)

   if timetables is not None:
      filtered_list = []
      for df_part in df_parts_list:
         is_tt = (model.part(obj_type = 'TimeSeries', related_uuid = model.uuid_for_part(df_part)) is not None)
         if timetables == is_tt:
            filtered_list.append(df_part)
      df_parts_list = filtered_list

   return df_parts_list


def timetable_parts_in_model(model, title = None, related_uuid = None):
   """Returns list of part names within model that are representing TimeTable dataframe support objects.

   arguments:
      model (model.Model): the model to be inspected for dataframes
      title (str, optional): if present, only parts with a citation title exactly matching will be
         included
      related_uuid (uuid, optional): if present, only parts relating to this uuid are included

   returns:
      list of str, each element in the list is a part name, within model, which is representing the support
      for a TimeTable object
   """

   return dataframe_parts_in_model(model, timetables = True, title = title, related_uuid = related_uuid)


def dataframe_for_title(model, title, realization = None):
   """Returns a DataFrame object loaded from model, with given title (optionally for given realization)."""

   df_parts = dataframe_parts_in_model(model, title = title)
   if df_parts is None or len(df_parts) == 0:
      return None
   assert len(df_parts) == 1
   return DataFrame(model, uuid = model.uuid_for_part(df_parts[0]), realization = realization)


def timetable_for_title(model, title, realization = None):
   """Returns a TimeTable object loaded from model, with given title (optionally for given realization)."""

   tt_parts = timetable_parts_in_model(model, title = title)
   if tt_parts is None or len(tt_parts) == 0:
      return None
   assert len(tt_parts) == 1
   return TimeTable(model, uuid = model.uuid_for_part(tt_parts[0]), realization = realization)



"""organize.py: RESQML Feature and Interpretation classes."""

version = '9th August 2021'

# most feature and interpretation classes catered for here
# stratigraphic classes in strata.py

import logging

log = logging.getLogger(__name__)
log.debug('organize.py version ' + version)

# import xml.etree.ElementTree as et
# from lxml import etree as et

import math as maths
import warnings

import resqpy.olio.xml_et as rqet
import resqpy.olio.uuid as bu
from resqpy.olio.xml_namespaces import curly_namespace as ns
from resqpy.olio.base import BaseResqpy


def extract_has_occurred_during(parent_node, tag = 'HasOccuredDuring'):  # RESQML Occured (stet)
   """Extracts UUIDs of chrono bottom and top from xml for has occurred during sub-node, or (None, None)."""
   hod_node = rqet.find_tag(parent_node, tag)
   if hod_node is None:
      return (None, None)
   else:
      return (rqet.find_nested_tags_text(hod_node, ['ChronoBottom', 'UUID']),
              rqet.find_nested_tags_text(hod_node, ['ChronoTop', 'UUID']))


def equivalent_chrono_pairs(pair_a, pair_b, model = None):
   if pair_a == pair_b:
      return True
   if pair_a is None or pair_b is None:
      return False
   if pair_a == (None, None) or pair_b == (None, None):
      return False
   if model is not None:
      # todo: compare chrono info by looking up xml based on the uuids
      pass
   return False  # cautious


def equivalent_extra_metadata(a, b):
   a_has = hasattr(a, 'extra_metadata')
   b_has = hasattr(b, 'extra_metadata')
   if a_has:
      a_em = a.extra_metadata
      a_has = len(a_em) > 0
   else:
      a_em = rqet.load_metadata_from_xml(a.root)
      a_has = a_em is not None and len(a_em) > 0
   if b_has:
      b_em = b.extra_metadata
      b_has = len(b_em) > 0
   else:
      b_em = rqet.load_metadata_from_xml(b.root)
      b_has = b_em is not None and len(b_em) > 0
   if a_has != b_has:
      return False
   if not a_has:
      return True
   return a_em == b_em


def create_xml_has_occurred_during(model, parent_node, hod_pair, tag = 'HasOccuredDuring'):
   if hod_pair is None:
      return
   base_chrono_uuid, top_chrono_uuid = hod_pair
   if base_chrono_uuid is None or top_chrono_uuid is None:
      return
   hod_node = rqet.SubElement(parent_node, tag)
   hod_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'TimeInterval')
   hod_node.text = rqet.null_xml_text
   chrono_base_root = model.root_for_uuid(base_chrono_uuid)
   chrono_top_root = model.root_for_uuid(top_chrono_uuid)
   model.create_ref_node('ChronoBottom', model.title_for_root(chrono_base_root), base_chrono_uuid, root = hod_node)
   model.create_ref_node('ChronoTop', model.title_for_root(chrono_top_root), top_chrono_uuid, root = hod_node)


def _alias_for_attribute(attribute_name):
   """Return an attribute that is a direct alias for an existing attribute"""

   def fget(self):
      return getattr(self, attribute_name)

   def fset(self, value):
      return setattr(self, attribute_name, value)

   return property(fget, fset, doc = f"Alias for {attribute_name}")


class OrganizationFeature(BaseResqpy):
   """Class for generic RESQML Organization Feature objects."""

   resqml_type = "OrganizationFeature"
   feature_name = _alias_for_attribute("title")

   def __init__(self,
                parent_model,
                root_node = None,
                uuid = None,
                feature_name = None,
                organization_kind = None,
                originator = None,
                extra_metadata = None):
      """Initialises an organization feature object."""

      self.organization_kind = organization_kind
      super().__init__(model = parent_model,
                       uuid = uuid,
                       title = feature_name,
                       originator = originator,
                       extra_metadata = extra_metadata,
                       root_node = root_node)

   def is_equivalent(self, other, check_extra_metadata = True):
      """Returns True if this feature is essentially the same as the other; otherwise False."""

      if not isinstance(other, OrganizationFeature):
         return False
      if self is other or bu.matching_uuids(self.uuid, other.uuid):
         return True
      return (self.feature_name == other.feature_name and self.organization_kind == other.organization_kind and
              ((not check_extra_metadata) or equivalent_extra_metadata(self, other)))

   def _load_from_xml(self):
      self.organization_kind = rqet.find_tag_text(self.root, 'OrganizationKind')

   def create_xml(self, add_as_part = True, originator = None, reuse = True):
      """Creates an organization feature xml node from this organization feature object."""

      if reuse and self.try_reuse():
         return self.root  # check for reusable (equivalent) object
      # create node with citation block
      ofn = super().create_xml(add_as_part = False, originator = originator)

      # Extra element for organization_kind
      if self.organization_kind not in ['earth model', 'fluid', 'stratigraphic', 'structural']:
         raise ValueError(self.organization_kind)
      kind_node = rqet.SubElement(ofn, ns['resqml2'] + 'OrganizationKind')
      kind_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'OrganizationKind')
      kind_node.text = self.organization_kind

      if add_as_part:
         self.model.add_part('obj_OrganizationFeature', self.uuid, ofn)

      return ofn


class GeobodyFeature(BaseResqpy):
   """Class for RESQML Geobody Feature objects (note: definition may be incomplete in RESQML 2.0.1)."""

   resqml_type = "GeobodyFeature"
   feature_name = _alias_for_attribute("title")

   def __init__(self, parent_model, root_node = None, uuid = None, feature_name = None, extra_metadata = None):
      """Initialises a geobody feature object."""

      super().__init__(model = parent_model,
                       uuid = uuid,
                       title = feature_name,
                       extra_metadata = extra_metadata,
                       root_node = root_node)

   def is_equivalent(self, other, check_extra_metadata = True):
      """Returns True if this feature is essentially the same as the other; otherwise False."""

      if other is None or not isinstance(other, self.__class__):
         return False
      if self is other or bu.matching_uuids(self.uuid, other.uuid):
         return True
      if check_extra_metadata and not equivalent_extra_metadata(self, other):
         return False
      return self.feature_name == other.feature_name

   def create_xml(self, add_as_part = True, originator = None, reuse = True):
      """Creates a geobody feature xml node from this geobody feature object."""
      if reuse and self.try_reuse():
         return self.root  # check for reusable (equivalent) object
      return super().create_xml(add_as_part = add_as_part, originator = originator)


class BoundaryFeature(BaseResqpy):
   """Class for RESQML Boudary Feature organizational objects."""

   resqml_type = "BoundaryFeature"
   feature_name = _alias_for_attribute("title")

   def __init__(self, parent_model, root_node = None, uuid = None, feature_name = None, extra_metadata = None):
      """Initialises a boundary feature organisational object."""

      super().__init__(model = parent_model,
                       uuid = uuid,
                       title = feature_name,
                       extra_metadata = extra_metadata,
                       root_node = root_node)

   def is_equivalent(self, other, check_extra_metadata = True):
      """Returns True if this feature is essentially the same as the other; otherwise False."""

      if other is None or not isinstance(other, BoundaryFeature):
         return False
      if self is other or bu.matching_uuids(self.uuid, other.uuid):
         return True
      if check_extra_metadata and not equivalent_extra_metadata(self, other):
         return False
      return self.feature_name == other.feature_name

   def create_xml(self, add_as_part = True, originator = None, reuse = True):
      """Creates a geobody feature xml node from this geobody feature object."""
      if reuse and self.try_reuse():
         return self.root  # check for reusable (equivalent) object
      return super().create_xml(add_as_part = add_as_part, originator = originator)


class FrontierFeature(BaseResqpy):
   """Class for RESQML Frontier Feature organizational objects."""

   resqml_type = "FrontierFeature"
   feature_name = _alias_for_attribute("title")

   def __init__(self, parent_model, root_node = None, uuid = None, feature_name = None, extra_metadata = None):
      """Initialises a frontier feature organisational object."""

      super().__init__(model = parent_model,
                       uuid = uuid,
                       title = feature_name,
                       extra_metadata = extra_metadata,
                       root_node = root_node)

   def is_equivalent(self, other, check_extra_metadata = True):
      """Returns True if this feature is essentially the same as the other; otherwise False."""

      if other is None or not isinstance(other, FrontierFeature):
         return False
      if self is other or bu.matching_uuids(self.uuid, other.uuid):
         return True
      if check_extra_metadata and not equivalent_extra_metadata(self, other):
         return False
      return self.feature_name == other.feature_name

   def create_xml(self, add_as_part = True, originator = None, reuse = True):
      """Creates a frontier feature organisational xml node from this frontier feature object."""
      if reuse and self.try_reuse():
         return self.root  # check for reusable (equivalent) object
      return super().create_xml(add_as_part = add_as_part, originator = originator)


class GeologicUnitFeature(BaseResqpy):
   """Class for RESQML Geologic Unit Feature organizational objects."""

   resqml_type = "GeologicUnitFeature"
   feature_name = _alias_for_attribute("title")

   def __init__(self, parent_model, root_node = None, uuid = None, feature_name = None, extra_metadata = None):
      """Initialises a geologic unit feature organisational object."""

      super().__init__(model = parent_model,
                       uuid = uuid,
                       title = feature_name,
                       extra_metadata = extra_metadata,
                       root_node = root_node)

   def is_equivalent(self, other, check_extra_metadata = True):
      """Returns True if this feature is essentially the same as the other; otherwise False."""

      if other is None or not isinstance(other, GeologicUnitFeature):
         return False
      if self is other or bu.matching_uuids(self.uuid, other.uuid):
         return True
      if check_extra_metadata and not equivalent_extra_metadata(self, other):
         return False
      return self.feature_name == other.feature_name

   def create_xml(self, add_as_part = True, originator = None, reuse = True):
      """Creates a geologic unit feature organisational xml node from this geologic unit feature object."""
      if reuse and self.try_reuse():
         return self.root  # check for reusable (equivalent) object
      return super().create_xml(add_as_part = add_as_part, originator = originator)


class FluidBoundaryFeature(BaseResqpy):
   """Class for RESQML Fluid Boundary Feature (contact) organizational objects."""

   resqml_type = "FluidBoundaryFeature"
   feature_name = _alias_for_attribute("title")
   valid_kinds = ('free water contact', 'gas oil contact', 'gas water contact', 'seal', 'water oil contact')

   def __init__(self,
                parent_model,
                root_node = None,
                uuid = None,
                kind = None,
                feature_name = None,
                extra_metadata = None):
      """Initialises a fluid boundary feature (contact) organisational object."""

      self.kind = kind
      super().__init__(model = parent_model,
                       uuid = uuid,
                       title = feature_name,
                       extra_metadata = extra_metadata,
                       root_node = root_node)

   def is_equivalent(self, other, check_extra_metadata = True):
      """Returns True if this feature is essentially the same as the other; otherwise False."""

      if other is None or not isinstance(other, FluidBoundaryFeature):
         return False
      if self is other or bu.matching_uuids(self.uuid, other.uuid):
         return True
      if check_extra_metadata and not equivalent_extra_metadata(self, other):
         return False
      return self.feature_name == other.feature_name and self.kind == other.kind

   def _load_from_xml(self):
      self.kind = rqet.find_tag_text(self.root, 'FluidContact')

   def create_xml(self, add_as_part = True, originator = None, reuse = True):
      """Creates a fluid boundary feature organisational xml node from this fluid boundary feature object."""

      if reuse and self.try_reuse():
         return self.root  # check for reusable (equivalent) object
      # create node with citation block
      fbf = super().create_xml(add_as_part = False, originator = originator)

      # Extra element for kind
      if self.kind not in self.valid_kinds:
         raise ValueError(f"fluid boundary feature kind '{self.kind}' not recognized")

      kind_node = rqet.SubElement(fbf, ns['resqml2'] + 'FluidContact')
      kind_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'FluidContact')
      kind_node.text = self.kind

      if add_as_part:
         self.model.add_part('obj_FluidBoundaryFeature', self.uuid, fbf)

      return fbf


class RockFluidUnitFeature(BaseResqpy):
   """Class for RESQML Rock Fluid Unit Feature organizational objects."""

   resqml_type = "RockFluidUnitFeature"
   feature_name = _alias_for_attribute("title")
   valid_phases = ('aquifer', 'gas cap', 'oil column', 'seal')

   def __init__(self,
                parent_model,
                root_node = None,
                uuid = None,
                phase = None,
                feature_name = None,
                top_boundary_feature = None,
                base_boundary_feature = None,
                extra_metadata = None):
      """Initialises a rock fluid unit feature organisational object."""

      self.phase = phase
      self.top_boundary_feature = top_boundary_feature
      self.base_boundary_feature = base_boundary_feature

      super().__init__(model = parent_model,
                       uuid = uuid,
                       title = feature_name,
                       extra_metadata = extra_metadata,
                       root_node = root_node)

   def is_equivalent(self, other, check_extra_metadata = True):
      """Returns True if this feature is essentially the same as the other; otherwise False."""

      if other is None or not isinstance(other, RockFluidUnitFeature):
         return False
      if self is other or bu.matching_uuids(self.uuid, other.uuid):
         return True
      if self.feature_name != other.feature_name or self.phase != other.phase:
         return False
      if self.top_boundary_feature is not None:
         if not self.top_boundary_feature.is_equivalent(other.top_boundary_feature):
            return False
      elif other.top_boundary_feature is not None:
         return False
      if self.base_boundary_feature is not None:
         if not self.base_boundary_feature.is_equivalent(other.base_boundary_feature):
            return False
      elif other.base_boundary_feature is not None:
         return False
      if check_extra_metadata and not equivalent_extra_metadata(self, other):
         return False
      return True

   def _load_from_xml(self):

      self.phase = rqet.find_tag_text(self.root, 'Phase')

      feature_ref_node = rqet.find_tag(self.root, 'FluidBoundaryTop')
      assert feature_ref_node is not None
      feature_root = self.model.referenced_node(feature_ref_node)
      feature_uuid = rqet.uuid_for_part_root(feature_root)
      assert feature_uuid is not None, 'rock fluid top boundary feature missing from model'
      self.top_boundary_feature = BoundaryFeature(self.model, uuid = feature_uuid)

      feature_ref_node = rqet.find_tag(self.root, 'FluidBoundaryBottom')
      assert feature_ref_node is not None
      feature_root = self.model.referenced_node(feature_ref_node)
      feature_uuid = rqet.uuid_for_part_root(feature_root)
      assert feature_uuid is not None, 'rock fluid bottom boundary feature missing from model'
      self.base_boundary_feature = BoundaryFeature(self.model, uuid = feature_uuid)

   def create_xml(self, add_as_part = True, add_relationships = True, originator = None, reuse = True):
      """Creates a rock fluid unit feature organisational xml node from this rock fluid unit feature object."""

      assert self.feature_name and self.phase and self.top_boundary_feature and self.base_boundary_feature
      if self.phase not in self.valid_phases:
         raise ValueError(f"Phase '{self.phase}' not recognized")

      if reuse and self.try_reuse():
         return self.root  # check for reusable (equivalent) object
      # create node with citation block
      rfuf = super().create_xml(add_as_part = False, originator = originator)

      phase_node = rqet.SubElement(rfuf, ns['resqml2'] + 'Phase')
      phase_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Phase')
      phase_node.text = self.phase

      top_boundary_root = self.top_boundary_feature.root
      assert top_boundary_root is not None
      self.model.create_ref_node('FluidBoundaryTop',
                                 self.model.title_for_root(top_boundary_root),
                                 top_boundary_root.attrib['uuid'],
                                 content_type = 'obj_BoundaryFeature',
                                 root = rfuf)

      base_boundary_root = self.base_boundary_feature.root
      assert base_boundary_root is not None
      self.model.create_ref_node('FluidBoundaryBottom',
                                 self.model.title_for_root(base_boundary_root),
                                 base_boundary_root.attrib['uuid'],
                                 content_type = 'obj_BoundaryFeature',
                                 root = rfuf)

      if add_as_part:
         self.model.add_part('obj_RockFluidUnitFeature', self.uuid, rfuf)
         if add_relationships:
            self.model.create_reciprocal_relationship(rfuf, 'destinationObject', top_boundary_root, 'sourceObject')
            self.model.create_reciprocal_relationship(rfuf, 'destinationObject', base_boundary_root, 'sourceObject')

      return rfuf


class TectonicBoundaryFeature(BaseResqpy):
   """Class for RESQML Tectonic Boundary Feature (fault) organizational objects."""

   resqml_type = "TectonicBoundaryFeature"
   feature_name = _alias_for_attribute("title")
   valid_kinds = ('fault', 'fracture')

   def __init__(self,
                parent_model,
                root_node = None,
                uuid = None,
                kind = None,
                feature_name = None,
                extra_metadata = None):
      """Initialises a tectonic boundary feature (fault or fracture) organisational object."""
      self.kind = kind
      super().__init__(model = parent_model,
                       uuid = uuid,
                       title = feature_name,
                       extra_metadata = extra_metadata,
                       root_node = root_node)

   def _load_from_xml(self):
      self.kind = rqet.find_tag_text(self.root, 'TectonicBoundaryKind')
      if not self.kind:
         self.kind = 'fault'

   def is_equivalent(self, other, check_extra_metadata = True):
      """Returns True if this feature is essentially the same as the other; otherwise False."""
      if other is None or not isinstance(other, TectonicBoundaryFeature):
         return False
      if self is other or bu.matching_uuids(self.uuid, other.uuid):
         return True
      if check_extra_metadata and not equivalent_extra_metadata(self, other):
         return False
      return self.feature_name == other.feature_name and self.kind == other.kind

   def create_xml(self, add_as_part = True, originator = None, reuse = True):
      """Creates a tectonic boundary feature organisational xml node from this tectonic boundary feature object."""

      if reuse and self.try_reuse():
         return self.root  # check for reusable (equivalent) object
      # create node with citation block
      tbf = super().create_xml(add_as_part = False, originator = originator)

      assert self.kind in self.valid_kinds
      kind_node = rqet.SubElement(tbf, ns['resqml2'] + 'TectonicBoundaryKind')
      kind_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'TectonicBoundaryKind')
      kind_node.text = self.kind

      if add_as_part:
         self.model.add_part('obj_TectonicBoundaryFeature', self.uuid, tbf)

      return tbf


class GeneticBoundaryFeature(BaseResqpy):
   """Class for RESQML Genetic Boundary Feature (horizon) organizational objects."""

   resqml_type = "GeneticBoundaryFeature"
   feature_name = _alias_for_attribute("title")
   valid_kinds = ('horizon', 'geobody boundary')

   def __init__(self,
                parent_model,
                root_node = None,
                uuid = None,
                kind = None,
                feature_name = None,
                extra_metadata = None):
      """Initialises a genetic boundary feature (horizon or geobody boundary) organisational object."""
      self.kind = kind
      self.absolute_age = None  # (timestamp, year offset) pair, or None; todo: support setting from args
      super().__init__(model = parent_model,
                       uuid = uuid,
                       title = feature_name,
                       extra_metadata = extra_metadata,
                       root_node = root_node)

   def _load_from_xml(self):
      self.kind = rqet.find_tag_text(self.root, 'GeneticBoundaryKind')
      age_node = rqet.find_tag(self.root, 'AbsoluteAge')
      if age_node:
         self.absolute_age = (rqet.find_tag_text(age_node, 'DateTime'), rqet.find_tag_int(age_node, 'YearOffset')
                             )  # year offset may be None

   def is_equivalent(self, other, check_extra_metadata = True):
      """Returns True if this feature is essentially the same as the other; otherwise False."""
      if other is None or not isinstance(other, GeneticBoundaryFeature):
         return False
      if self is other or bu.matching_uuids(self.uuid, other.uuid):
         return True
      if check_extra_metadata and not equivalent_extra_metadata(self, other):
         return False
      return self.feature_name == other.feature_name and self.kind == other.kind and self.absolute_age == other.absolute_age

   def create_xml(self, add_as_part = True, originator = None, reuse = True):
      """Creates a genetic boundary feature organisational xml node from this genetic boundary feature object."""

      if reuse and self.try_reuse():
         return self.root  # check for reusable (equivalent) object
      # create node with citation block
      gbf = super().create_xml(add_as_part = False, originator = originator)

      assert self.kind in self.valid_kinds
      kind_node = rqet.SubElement(gbf, ns['resqml2'] + 'GeneticBoundaryKind')
      kind_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'GeneticBoundaryKind')
      kind_node.text = self.kind

      if self.absolute_age is not None:
         date_time, year_offset = self.absolute_age
         age_node = rqet.SubElement(gbf, ns['resqml2'] + 'AbsoluteAge')
         age_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Timestamp')
         age_node.text = rqet.null_xml_text
         dt_node = rqet.SubElement(age_node, ns['resqml2'] + 'DateTime')
         dt_node.set(ns['xsi'] + 'type', ns['xsd'] + 'dateTime')
         dt_node.text = str(date_time)
         if year_offset is not None:
            yo_node = rqet.SubElement(age_node, ns['resqml2'] + 'YearOffset')
            yo_node.set(ns['xsi'] + 'type', ns['xsd'] + 'long')
            yo_node.text = str(year_offset)

      if add_as_part:
         self.model.add_part('obj_GeneticBoundaryFeature', self.uuid, gbf)

      return gbf


class WellboreFeature(BaseResqpy):
   """Class for RESQML Wellbore Feature organizational objects."""

   # note: optional WITSML link not supported

   resqml_type = "WellboreFeature"
   feature_name = _alias_for_attribute("title")

   def __init__(self, parent_model, root_node = None, uuid = None, feature_name = None, extra_metadata = None):
      """Initialises a wellbore feature organisational object."""
      super().__init__(model = parent_model,
                       uuid = uuid,
                       title = feature_name,
                       extra_metadata = extra_metadata,
                       root_node = root_node)

   def is_equivalent(self, other, check_extra_metadata = True):
      """Returns True if this feature is essentially the same as the other; otherwise False."""
      if other is None or not isinstance(other, WellboreFeature):
         return False
      if self is other or bu.matching_uuids(self.uuid, other.uuid):
         return True
      if check_extra_metadata and not equivalent_extra_metadata(self, other):
         return False
      return self.feature_name == other.feature_name

   def create_xml(self, add_as_part = True, originator = None, reuse = True):
      """Creates a wellbore feature organisational xml node from this wellbore feature object."""
      if reuse and self.try_reuse():
         return self.root  # check for reusable (equivalent) object
      return super().create_xml(add_as_part = add_as_part, originator = originator)


class FaultInterpretation(BaseResqpy):
   """Class for RESQML Fault Interpretation organizational objects.

   RESQML documentation:

      A type of boundary feature, this class contains the data describing an opinion
      about the characterization of the fault, which includes the attributes listed below.

   """

   resqml_type = "FaultInterpretation"
   valid_domains = ('depth', 'time', 'mixed')

   # note: many of the attributes could be deduced from geometry

   def __init__(self,
                parent_model,
                root_node = None,
                uuid = None,
                title = None,
                tectonic_boundary_feature = None,
                domain = 'depth',
                is_normal = None,
                is_listric = None,
                maximum_throw = None,
                mean_azimuth = None,
                mean_dip = None,
                extra_metadata = None):
      """Initialises a Fault interpretation organisational object."""

      # note: will create a paired TectonicBoundaryFeature object when loading from xml
      # if not extracting from xml,:
      # tectonic_boundary_feature is required and must be a TectonicBoundaryFeature object
      # domain is required and must be one of 'depth', 'time' or 'mixed'
      # is_listric is required if the fault is not normal (and must be None if normal)
      # max throw, azimuth & dip are all optional
      # the throw interpretation list is not supported for direct initialisation

      self.tectonic_boundary_feature = tectonic_boundary_feature  # InterpretedFeature RESQML field, when not loading from xml
      self.feature_root = None if self.tectonic_boundary_feature is None else self.tectonic_boundary_feature.root
      if (not title) and self.tectonic_boundary_feature is not None:
         title = self.tectonic_boundary_feature.feature_name
      self.main_has_occurred_during = (None, None)
      self.is_normal = is_normal  # extra field, not explicitly in RESQML
      self.domain = domain
      # RESQML xml business rule: IsListric must be present if the fault is normal; must not be present if the fault is not normal
      self.is_listric = is_listric
      self.maximum_throw = maximum_throw
      self.mean_azimuth = mean_azimuth
      self.mean_dip = mean_dip
      self.throw_interpretation_list = None  # list of (list of throw kind, (base chrono uuid, top chrono uuid)))

      super().__init__(model = parent_model,
                       uuid = uuid,
                       title = title,
                       extra_metadata = extra_metadata,
                       root_node = root_node)

   @property
   def feature_uuid(self):
      # TODO: rewrite using uuid as primary key
      return rqet.uuid_for_part_root(self.feature_root)

   def _load_from_xml(self):
      root_node = self.root
      self.domain = rqet.find_tag_text(root_node, 'Domain')
      interp_feature_ref_node = rqet.find_tag(root_node, 'InterpretedFeature')
      assert interp_feature_ref_node is not None
      self.feature_root = self.model.referenced_node(interp_feature_ref_node)
      if self.feature_root is not None:
         self.tectonic_boundary_feature = TectonicBoundaryFeature(self.model,
                                                                  uuid = self.feature_root.attrib['uuid'],
                                                                  feature_name = self.model.title_for_root(
                                                                     self.feature_root))
         self.main_has_occurred_during = extract_has_occurred_during(root_node)
         self.is_listric = rqet.find_tag_bool(root_node, 'IsListric')
         self.is_normal = (self.is_listric is None)
         self.maximum_throw = rqet.find_tag_float(root_node, 'MaximumThrow')
         # todo: check that type="eml:LengthMeasure" is simple float
         self.mean_azimuth = rqet.find_tag_float(root_node, 'MeanAzimuth')
         self.mean_dip = rqet.find_tag_float(root_node, 'MeanDip')
         throw_interpretation_nodes = rqet.list_of_tag(root_node, 'ThrowInterpretation')
         if throw_interpretation_nodes is not None and len(throw_interpretation_nodes):
            self.throw_interpretation_list = []
            for ti_node in throw_interpretation_nodes:
               hod_pair = extract_has_occurred_during(ti_node)
               throw_kind_list = rqet.list_of_tag(ti_node, 'Throw')
               for tk_node in throw_kind_list:
                  self.throw_interpretation_list.append((tk_node.text, hod_pair))

   def is_equivalent(self, other, check_extra_metadata = True):
      """Returns True if this interpretation is essentially the same as the other; otherwise False."""

      if other is None or not isinstance(other, FaultInterpretation):
         return False
      if self is other or bu.matching_uuids(self.uuid, other.uuid):
         return True
      if self.tectonic_boundary_feature is not None:
         if not self.tectonic_boundary_feature.is_equivalent(other.tectonic_boundary_feature):
            return False
      elif other.tectonic_boundary_feature is not None:
         return False
      if self.root is not None and other.root is not None:
         if rqet.citation_title_for_node(self.root) != rqet.citation_title_for_node(other.root):
            return False
      elif self.root is not None or other.root is not None:
         return False
      if (not equivalent_chrono_pairs(self.main_has_occurred_during, other.main_has_occurred_during) or
          self.is_normal != other.is_normal or self.domain != other.domain or self.is_listric != other.is_listric):
         return False
      if ((self.maximum_throw is None) != (other.maximum_throw is None) or (self.mean_azimuth is None) !=
          (other.mean_azimuth is None) or (self.mean_dip is None) != (other.mean_dip is None)):
         return False
      if self.maximum_throw is not None and not maths.isclose(self.maximum_throw, other.maximum_throw, rel_tol = 1e-3):
         return False
      if self.mean_azimuth is not None and not maths.isclose(self.mean_azimuth, other.mean_azimuth, abs_tol = 0.5):
         return False
      if self.mean_dip is not None and not maths.isclose(self.mean_dip, other.mean_dip, abs_tol = 0.5):
         return False
      if check_extra_metadata and not equivalent_extra_metadata(self, other):
         return False
      if not self.throw_interpretation_list and not other.throw_interpretation_list:
         return True
      if not self.throw_interpretation_list or not other.throw_interpretation_list:
         return False
      if len(self.throw_interpretation_list) != len(other.throw_interpretation_list):
         return False
      for this_ti, other_ti in zip(self.throw_interpretation_list, other.throw_interpretation_list):
         if this_ti[0] != other_ti[0]:
            return False  # throw kind
         if not equivalent_chrono_pairs(this_ti[1], other_ti[1]):
            return False
      return True

   def create_xml(self,
                  tectonic_boundary_feature_root = None,
                  add_as_part = True,
                  add_relationships = True,
                  originator = None,
                  title_suffix = None,
                  reuse = True):
      """Creates a fault interpretation organisational xml node from a fault interpretation object."""

      # note: related tectonic boundary feature node should be created first and referenced here

      assert self.is_normal == (self.is_listric is None)
      if not self.title:
         if tectonic_boundary_feature_root is not None:
            title = rqet.find_nested_tags_text(tectonic_boundary_feature_root, ['Citation', 'Title'])
         else:
            title = 'fault interpretation'
         if title_suffix:
            title += ' ' + title_suffix

      if reuse and self.try_reuse():
         return self.root

      fi = super().create_xml(add_as_part = False, originator = originator)

      if self.tectonic_boundary_feature is not None:
         tbf_root = self.tectonic_boundary_feature.root
         if tbf_root is not None:
            if tectonic_boundary_feature_root is None:
               tectonic_boundary_feature_root = tbf_root
            else:
               assert tbf_root is tectonic_boundary_feature_root, 'tectonic boundary feature mismatch'
      assert tectonic_boundary_feature_root is not None

      assert self.domain in self.valid_domains, 'illegal domain value for fault interpretation'
      dom_node = rqet.SubElement(fi, ns['resqml2'] + 'Domain')
      dom_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Domain')
      dom_node.text = self.domain

      self.model.create_ref_node('InterpretedFeature',
                                 self.model.title_for_root(tectonic_boundary_feature_root),
                                 tectonic_boundary_feature_root.attrib['uuid'],
                                 content_type = 'obj_TectonicBoundaryFeature',
                                 root = fi)

      create_xml_has_occurred_during(self.model, fi, self.main_has_occurred_during)

      if self.is_listric is not None:
         listric = rqet.SubElement(fi, ns['resqml2'] + 'IsListric')
         listric.set(ns['xsi'] + 'type', ns['xsd'] + 'boolean')
         listric.text = str(self.is_listric).lower()

      # todo: check eml:LengthMeasure and PlaneAngleMeasure structures: uom?
      if self.maximum_throw is not None:
         max_throw = rqet.SubElement(fi, ns['resqml2'] + 'MaximumThrow')
         max_throw.set(ns['xsi'] + 'type', ns['eml'] + 'LengthMeasure')
         max_throw.text = str(self.maximum_throw)

      if self.mean_azimuth is not None:
         azimuth = rqet.SubElement(fi, ns['resqml2'] + 'MeanAzimuth')
         azimuth.set(ns['xsi'] + 'type', ns['eml'] + 'PlaneAngleMeasure')
         azimuth.text = str(self.mean_azimuth)

      if self.mean_dip is not None:
         dip = rqet.SubElement(fi, ns['resqml2'] + 'MeanDip')
         dip.set(ns['xsi'] + 'type', ns['eml'] + 'PlaneAngleMeasure')
         dip.text = str(self.mean_azimuth)

      if self.throw_interpretation_list is not None and len(self.throw_interpretation_list):
         previous_has_occurred_during = ('never', 'never')
         ti_node = None
         for (throw_kind, (base_chrono_uuid, top_chrono_uuid)) in self.throw_interpretation_list:
            if (str(base_chrono_uuid), str(top_chrono_uuid)) != previous_has_occurred_during:
               previous_has_occurred_during = (str(base_chrono_uuid), str(top_chrono_uuid))
               ti_node = rqet.SubElement(fi, 'ThrowInterpretation')
               ti_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'FaultThrow')
               ti_node.text = rqet.null_xml_text
               create_xml_has_occurred_during(self.model, ti_node, (base_chrono_uuid, top_chrono_uuid))
            tk_node = rqet.SubElement(ti_node, 'Throw')
            tk_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'ThrowKind')
            tk_node.text = throw_kind

      if add_as_part:
         self.model.add_part('obj_FaultInterpretation', self.uuid, fi)
         if add_relationships:
            self.model.create_reciprocal_relationship(fi, 'destinationObject', tectonic_boundary_feature_root,
                                                      'sourceObject')

      return fi


class EarthModelInterpretation(BaseResqpy):
   """Class for RESQML Earth Model Interpretation organizational objects."""

   # TODO: add support for StratigraphicColumn reference and other optional references

   resqml_type = 'EarthModelInterpretation'
   valid_domains = ('depth', 'time', 'mixed')

   def __init__(self,
                parent_model,
                root_node = None,
                uuid = None,
                title = None,
                organization_feature = None,
                domain = 'depth',
                extra_metadata = None):
      """Initialises an earth model interpretation organisational object."""
      self.domain = domain
      self.organization_feature = organization_feature  # InterpretedFeature RESQML field
      self.feature_root = None if self.organization_feature is None else self.organization_feature.root
      self.has_occurred_during = (None, None)
      if (not title) and organization_feature is not None:
         title = organization_feature.feature_name
      super().__init__(model = parent_model,
                       uuid = uuid,
                       title = title,
                       extra_metadata = extra_metadata,
                       root_node = root_node)

   def _load_from_xml(self):
      self.domain = rqet.find_tag_text(self.root, 'Domain')
      interp_feature_ref_node = rqet.find_tag(self.root, 'InterpretedFeature')
      assert interp_feature_ref_node is not None
      self.feature_root = self.model.referenced_node(interp_feature_ref_node)
      if self.feature_root is not None:
         self.organization_feature = OrganizationFeature(self.model,
                                                         uuid = self.feature_root.attrib['uuid'],
                                                         feature_name = self.model.title_for_root(self.feature_root))
      self.has_occurred_during = extract_has_occurred_during(self.root)

   def is_equivalent(self, other, check_extra_metadata = True):
      """Returns True if this interpretation is essentially the same as the other; otherwise False."""
      if other is None or not isinstance(other, EarthModelInterpretation):
         return False
      if self is other or bu.matching_uuids(self.uuid, other.uuid):
         return True
      if self.organization_feature is not None:
         if not self.organization_feature.is_equivalent(other.organization_feature):
            return False
      elif other.organization_feature is not None:
         return False
      if self.root is not None and other.root is not None:
         if rqet.citation_title_for_node(self.root) != rqet.citation_title_for_node(other.root):
            return False
      elif self.root is not None or other.root is not None:
         return False
      if check_extra_metadata and not equivalent_extra_metadata(self, other):
         return False
      return self.domain == other.domain and equivalent_chrono_pairs(self.has_occurred_during,
                                                                     other.has_occurred_during)

   def create_xml(self,
                  organization_feature_root = None,
                  add_as_part = True,
                  add_relationships = True,
                  originator = None,
                  title_suffix = None,
                  reuse = True):
      """Creates an earth model interpretation organisational xml node from an earth model interpretation object."""

      # note: related organization feature node should be created first and referenced here

      if not self.title:
         self.title = self.organization_feature.feature_name
      if title_suffix:
         self.title += ' ' + title_suffix

      if reuse and self.try_reuse():
         return self.root
      emi = super().create_xml(add_as_part = False, originator = originator)

      if self.organization_feature is not None:
         of_root = self.organization_feature.root
         if of_root is not None:
            if organization_feature_root is None:
               organization_feature_root = of_root
            else:
               assert of_root is organization_feature_root, 'organization feature mismatch'

      assert organization_feature_root is not None, 'interpreted feature not established for model interpretation'

      assert self.domain in self.valid_domains, 'illegal domain value for earth model interpretation'
      dom_node = rqet.SubElement(emi, ns['resqml2'] + 'Domain')
      dom_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Domain')
      dom_node.text = self.domain

      self.model.create_ref_node('InterpretedFeature',
                                 self.model.title_for_root(organization_feature_root),
                                 organization_feature_root.attrib['uuid'],
                                 content_type = 'obj_OrganizationFeature',
                                 root = emi)

      create_xml_has_occurred_during(self.model, emi, self.has_occurred_during)

      if add_as_part:
         self.model.add_part('obj_EarthModelInterpretation', self.uuid, emi)
         if add_relationships:
            self.model.create_reciprocal_relationship(emi, 'destinationObject', organization_feature_root,
                                                      'sourceObject')

      return emi


class HorizonInterpretation(BaseResqpy):
   """Class for RESQML Horizon Interpretation organizational objects."""

   resqml_type = 'HorizonInterpretation'
   valid_domains = ('depth', 'time', 'mixed')
   valid_sequence_stratigraphy_surfaces = ('flooding', 'ravinement', 'maximum flooding', 'transgressive')
   valid_boundary_relations = ('conformable', 'unconformable below and above', 'unconformable above',
                               'unconformable below')

   def __init__(self,
                parent_model,
                root_node = None,
                uuid = None,
                title = None,
                genetic_boundary_feature = None,
                domain = 'depth',
                boundary_relation_list = None,
                sequence_stratigraphy_surface = None,
                extra_metadata = None):
      """Initialises a horizon interpretation organisational object."""

      # note: will create a paired GeneticBoundaryFeature object when loading from xml (and possibly a Surface object)

      self.domain = domain
      self.genetic_boundary_feature = genetic_boundary_feature  # InterpretedFeature RESQML field, when not loading from xml
      self.feature_root = None if self.genetic_boundary_feature is None else self.genetic_boundary_feature.root
      if (not title) and self.genetic_boundary_feature is not None:
         title = self.genetic_boundary_feature.feature_name
      self.has_occurred_during = (None, None)
      self.boundary_relation_list = None if not boundary_relation_list else boundary_relation_list.copy()
      self.sequence_stratigraphy_surface = sequence_stratigraphy_surface
      super().__init__(model = parent_model,
                       uuid = uuid,
                       title = title,
                       extra_metadata = extra_metadata,
                       root_node = root_node)

   def _load_from_xml(self):
      self.domain = rqet.find_tag_text(self.root, 'Domain')
      interp_feature_ref_node = rqet.find_tag(self.root, 'InterpretedFeature')
      assert interp_feature_ref_node is not None
      self.feature_root = self.model.referenced_node(interp_feature_ref_node)
      if self.feature_root is not None:
         self.genetic_boundary_feature = GeneticBoundaryFeature(self.model,
                                                                kind = 'horizon',
                                                                uuid = self.feature_root.attrib['uuid'],
                                                                feature_name = self.model.title_for_root(
                                                                   self.feature_root))
      self.has_occurred_during = extract_has_occurred_during(self.root)
      br_node_list = rqet.list_of_tag(self.root, 'BoundaryRelation')
      if br_node_list is not None and len(br_node_list) > 0:
         self.boundary_relation_list = []
         for br_node in br_node_list:
            self.boundary_relation_list.append(br_node.text)
      self.sequence_stratigraphy_surface = rqet.find_tag_text(self.root, 'SequenceStratigraphySurface')

   def is_equivalent(self, other, check_extra_metadata = True):
      """Returns True if this interpretation is essentially the same as the other; otherwise False."""

      if other is None or not isinstance(other, HorizonInterpretation):
         return False
      if self is other or bu.matching_uuids(self.uuid, other.uuid):
         return True
      if self.genetic_boundary_feature is not None:
         if not self.genetic_boundary_feature.is_equivalent(other.genetic_boundary_feature):
            return False
      elif other.genetic_boundary_feature is not None:
         return False
      if self.root is not None and other.root is not None:
         if rqet.citation_title_for_node(self.root) != rqet.citation_title_for_node(other.root):
            return False
      elif self.root is not None or other.root is not None:
         return False
      if (self.domain != other.domain or
          not equivalent_chrono_pairs(self.has_occurred_during, other.has_occurred_during) or
          self.sequence_stratigraphy_surface != other.sequence_stratigraphy_surface):
         return False
      if check_extra_metadata and not equivalent_extra_metadata(self, other):
         return False
      if not self.boundary_relation_list and not other.boundary_relation_list:
         return True
      if not self.boundary_relation_list or not other.boundary_relation_list:
         return False
      return set(self.boundary_relation_list) == set(other.boundary_relation_list)

   def create_xml(self,
                  genetic_boundary_feature_root = None,
                  add_as_part = True,
                  add_relationships = True,
                  originator = None,
                  title_suffix = None,
                  reuse = True):
      """Creates a horizon interpretation organisational xml node from a horizon interpretation object."""

      # note: related genetic boundary feature node should be created first and referenced here

      if not self.title:
         self.title = self.genetic_boundary_feature.feature_name
      if title_suffix:
         self.title += ' ' + title_suffix

      if reuse and self.try_reuse():
         return self.root
      hi = super().create_xml(add_as_part = False, originator = originator)

      if self.genetic_boundary_feature is not None:
         gbf_root = self.genetic_boundary_feature.root
         if gbf_root is not None:
            if genetic_boundary_feature_root is None:
               genetic_boundary_feature_root = gbf_root
            else:
               assert gbf_root is genetic_boundary_feature_root, 'genetic boundary feature mismatch'

      assert self.domain in self.valid_domains, 'illegal domain value for horizon interpretation'
      dom_node = rqet.SubElement(hi, ns['resqml2'] + 'Domain')
      dom_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Domain')
      dom_node.text = self.domain

      create_xml_has_occurred_during(self.model, hi, self.has_occurred_during)

      if self.boundary_relation_list is not None:
         for boundary_relation in self.boundary_relation_list:
            assert boundary_relation in self.valid_boundary_relations
            br_node = rqet.SubElement(hi, ns['resqml2'] + 'BoundaryRelation')
            br_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'BoundaryRelation')
            br_node.text = boundary_relation

      if self.sequence_stratigraphy_surface is not None:
         assert self.sequence_stratigraphy_surface in self.valid_sequence_stratigraphy_surfaces
         sss_node = rqet.SubElement(hi, ns['resqml2'] + 'SequenceStratigraphySurface')
         sss_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'SequenceStratigraphySurface')
         sss_node.text = self.sequence_stratigraphy_surface

      self.model.create_ref_node('InterpretedFeature',
                                 self.model.title_for_root(genetic_boundary_feature_root),
                                 genetic_boundary_feature_root.attrib['uuid'],
                                 content_type = 'obj_GeneticBoundaryFeature',
                                 root = hi)

      if add_as_part:
         self.model.add_part('obj_HorizonInterpretation', self.uuid, hi)
         if add_relationships:
            self.model.create_reciprocal_relationship(hi, 'destinationObject', genetic_boundary_feature_root,
                                                      'sourceObject')

      return hi


class GeobodyBoundaryInterpretation(BaseResqpy):
   """Class for RESQML Geobody Boudary Interpretation organizational objects."""

   resqml_type = 'GeobodyBoundaryInterpretation'
   valid_domains = ('depth', 'time', 'mixed')
   valid_boundary_relations = ('conformable', 'unconformable below and above', 'unconformable above',
                               'unconformable below')

   def __init__(self,
                parent_model,
                root_node = None,
                uuid = None,
                title = None,
                genetic_boundary_feature = None,
                domain = 'depth',
                boundary_relation_list = None,
                extra_metadata = None):

      self.domain = domain
      self.boundary_relation_list = None if not boundary_relation_list else boundary_relation_list.copy()
      self.genetic_boundary_feature = genetic_boundary_feature  # InterpretedFeature RESQML field, when not loading from xml
      self.feature_root = None if self.genetic_boundary_feature is None else self.genetic_boundary_feature.root
      if (not title) and self.genetic_boundary_feature is not None:
         title = self.genetic_boundary_feature.feature_name
      self.has_occurred_during = (None, None)
      super().__init__(model = parent_model,
                       uuid = uuid,
                       title = title,
                       extra_metadata = extra_metadata,
                       root_node = root_node)

   def _load_from_xml(self):
      self.domain = rqet.find_tag_text(self.root, 'Domain')
      interp_feature_ref_node = rqet.find_tag(self.root, 'InterpretedFeature')
      assert interp_feature_ref_node is not None
      self.feature_root = self.model.referenced_node(interp_feature_ref_node)
      if self.feature_root is not None:
         self.genetic_boundary_feature = GeneticBoundaryFeature(self.model,
                                                                kind = 'geobody boundary',
                                                                uuid = self.feature_root.attrib['uuid'],
                                                                feature_name = self.model.title_for_root(
                                                                   self.feature_root))
      self.has_occurred_during = extract_has_occurred_during(self.root)
      br_node_list = rqet.list_of_tag(self.root, 'BoundaryRelation')
      if br_node_list is not None and len(br_node_list) > 0:
         self.boundary_relation_list = []
         for br_node in br_node_list:
            self.boundary_relation_list.append(br_node.text)

   def is_equivalent(self, other, check_extra_metadata = True):
      """Returns True if this interpretation is essentially the same as the other; otherwise False."""

      if other is None or not isinstance(other, GeobodyBoundaryInterpretation):
         return False
      if self is other or bu.matching_uuids(self.uuid, other.uuid):
         return True
      if self.genetic_boundary_feature is not None:
         if not self.genetic_boundary_feature.is_equivalent(other.genetic_boundary_feature):
            return False
      elif other.genetic_boundary_feature is not None:
         return False
      if self.root is not None and other.root is not None:
         if rqet.citation_title_for_node(self.root) != rqet.citation_title_for_node(other.root):
            return False
      elif self.root is not None or other.root is not None:
         return False
      if (self.domain != other.domain or
          not equivalent_chrono_pairs(self.has_occurred_during, other.has_occurred_during)):
         return False
      if check_extra_metadata and not equivalent_extra_metadata(self, other):
         return False
      if not self.boundary_relation_list and not other.boundary_relation_list:
         return True
      if not self.boundary_relation_list or not other.boundary_relation_list:
         return False
      return set(self.boundary_relation_list) == set(other.boundary_relation_list)

   def create_xml(self,
                  genetic_boundary_feature_root = None,
                  add_as_part = True,
                  add_relationships = True,
                  originator = None,
                  title_suffix = None,
                  reuse = True):
      """Creates a geobody boundary interpretation organisational xml node from a geobody boundary interpretation object."""

      if not self.title:
         self.title = self.genetic_boundary_feature.feature_name
      if title_suffix:
         self.title += ' ' + title_suffix

      if reuse and self.try_reuse():
         return self.root
      gbi = super().create_xml(add_as_part = False, originator = originator)

      if self.genetic_boundary_feature is not None:
         gbf_root = self.genetic_boundary_feature.root
         if gbf_root is not None:
            if genetic_boundary_feature_root is None:
               genetic_boundary_feature_root = gbf_root
            else:
               assert gbf_root is genetic_boundary_feature_root, 'genetic boundary feature mismatch'
      else:
         if genetic_boundary_feature_root is None:
            genetic_boundary_feature_root = self.feature_root
         assert genetic_boundary_feature_root is not None
         self.genetic_boundary_feature = GeneticBoundaryFeature(self.model,
                                                                uuid = genetic_boundary_feature_root.attrib['uuid'])
      self.feature_root = genetic_boundary_feature_root

      assert self.domain in self.valid_domains, 'illegal domain value for geobody boundary interpretation'
      dom_node = rqet.SubElement(gbi, ns['resqml2'] + 'Domain')
      dom_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Domain')
      dom_node.text = self.domain

      create_xml_has_occurred_during(self.model, gbi, self.has_occurred_during)

      if self.boundary_relation_list is not None:
         for boundary_relation in self.boundary_relation_list:
            assert boundary_relation in self.valid_boundary_relations
            br_node = rqet.SubElement(gbi, ns['resqml2'] + 'BoundaryRelation')
            br_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'BoundaryRelation')
            br_node.text = str(boundary_relation)

      self.model.create_ref_node('InterpretedFeature',
                                 self.model.title_for_root(genetic_boundary_feature_root),
                                 genetic_boundary_feature_root.attrib['uuid'],
                                 content_type = 'obj_GeneticBoundaryFeature',
                                 root = gbi)

      if add_as_part:
         self.model.add_part('obj_GeobodyBoundaryInterpretation', self.uuid, gbi)
         if add_relationships:
            self.model.create_reciprocal_relationship(gbi, 'destinationObject', genetic_boundary_feature_root,
                                                      'sourceObject')

      return gbi


class GeobodyInterpretation(BaseResqpy):
   """Class for RESQML Geobody Interpretation objects."""

   resqml_type = 'GeobodyInterpretation'
   valid_domains = ('depth', 'time', 'mixed')
   valid_compositions = (
      'intrusive clay',
      'organic',
      'intrusive mud',
      'evaporite salt',
      'evaporite non salt',
      'sedimentary siliclastic',
      'carbonate',
      'magmatic intrusive granitoid',
      'magmatic intrusive pyroclastic',
      'magmatic extrusive lava flow',
      'other chemichal rock',  # chemichal (stet: from xsd)
      'other chemical rock',
      'sedimentary turbidite')
   valid_implacements = ('autochtonous', 'allochtonous')
   valid_geobody_shapes = ('dyke', 'silt', 'sill', 'dome', 'sheeth', 'sheet', 'diapir', 'batholith', 'channel', 'delta',
                           'dune', 'fan', 'reef', 'wedge')

   def __init__(self,
                parent_model,
                root_node = None,
                uuid = None,
                title = None,
                geobody_feature = None,
                domain = 'depth',
                composition = None,
                material_implacement = None,
                geobody_shape = None,
                extra_metadata = None):
      """Initialise a new geobody interpretation object, either from xml or explicitly."""

      self.domain = domain
      self.geobody_feature = geobody_feature  # InterpretedFeature RESQML field, when not loading from xml
      self.feature_root = None if self.geobody_feature is None else self.geobody_feature.root
      self.has_occurred_during = (None, None)
      self.composition = composition
      self.implacement = material_implacement
      self.geobody_shape = geobody_shape
      super().__init__(model = parent_model,
                       uuid = uuid,
                       title = title,
                       extra_metadata = extra_metadata,
                       root_node = root_node)

   def _load_from_xml(self):
      interp_feature_ref_node = rqet.find_tag(self.root, 'InterpretedFeature')
      assert interp_feature_ref_node is not None
      self.feature_root = self.model.referenced_node(interp_feature_ref_node)
      if self.feature_root is not None:
         self.geobody_feature = GeobodyFeature(self.model,
                                               uuid = self.feature_root.attrib['uuid'],
                                               feature_name = self.model.title_for_root(self.feature_root))
      self.has_occurred_during = extract_has_occurred_during(self.root)
      self.composition = rqet.find_tag_text(self.root, 'GeologicUnitComposition')
      self.implacement = rqet.find_tag_text(self.root, 'GeologicUnitMaterialImplacement')
      self.geobody_shape = rqet.find_tag_text(self.root, 'Geobody3dShape')

   def is_equivalent(self, other, check_extra_metadata = True):
      """Returns True if this interpretation is essentially the same as the other; otherwise False."""

      if other is None or not isinstance(other, GeobodyInterpretation):
         return False
      if self is other or bu.matching_uuids(self.uuid, other.uuid):
         return True
      if self.geobody_feature is not None:
         if not self.geobody_feature.is_equivalent(other.geobody_feature):
            return False
      elif other.geobody_feature is not None:
         return False
      if self.root is not None and other.root is not None:
         if rqet.citation_title_for_node(self.root) != rqet.citation_title_for_node(other.root):
            return False
      elif self.root is not None or other.root is not None:
         return False
      if check_extra_metadata and not equivalent_extra_metadata(self, other):
         return False
      return (self.domain == other.domain and
              equivalent_chrono_pairs(self.has_occurred_during, other.has_occurred_during) and
              self.composition == other.composition and self.implacement == other.implacement and
              self.geobody_shape == other.geobody_shape)

   def create_xml(self,
                  geobody_feature_root = None,
                  add_as_part = True,
                  add_relationships = True,
                  originator = None,
                  title_suffix = None,
                  reuse = True):

      if not self.title:
         self.title = self.geobody_feature.feature_name
      if title_suffix:
         self.title += ' ' + title_suffix

      if reuse and self.try_reuse():
         return self.root
      gi = super().create_xml(add_as_part = False, originator = originator)

      if self.geobody_feature is not None:
         gbf_root = self.geobody_feature.root
         if gbf_root is not None:
            if geobody_feature_root is None:
               geobody_feature_root = gbf_root
            else:
               assert gbf_root is geobody_feature_root, 'geobody feature mismatch'
      else:
         if geobody_feature_root is None:
            geobody_feature_root = self.feature_root
         assert geobody_feature_root is not None
         self.geobody_feature = GeobodyFeature(self.model, uuid = geobody_feature_root.attrib['uuid'])
      self.feature_root = geobody_feature_root

      assert self.domain in self.valid_domains, 'illegal domain value for geobody interpretation'
      dom_node = rqet.SubElement(gi, ns['resqml2'] + 'Domain')
      dom_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Domain')
      dom_node.text = self.domain

      create_xml_has_occurred_during(self.model, gi, self.has_occurred_during)

      if self.composition:
         assert self.composition in self.valid_compositions
         guc_node = rqet.SubElement(gi, ns['resqml2'] + 'GeologicUnitComposition')
         guc_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'GeologicUnitComposition')
         guc_node.text = self.composition
         # if self.composition.startswith('intrusive'): guc_node.text += ' '

      if self.implacement:
         assert self.implacement in self.valid_implacements
         gumi_node = rqet.SubElement(gi, ns['resqml2'] + 'GeologicUnitMaterialImplacement')
         gumi_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'GeologicUnitMaterialImplacement')
         gumi_node.text = self.implacement

      if self.geobody_shape:
         # note: 'silt' & 'sheeth' believed erroneous, so 'sill' and 'sheet' added
         assert self.geobody_shape in self.valid_geobody_shapes
         gs_node = rqet.SubElement(gi, ns['resqml2'] + 'Geobody3dShape')
         gs_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Geobody3dShape')
         gs_node.text = self.geobody_shape

      self.model.create_ref_node('InterpretedFeature',
                                 self.model.title_for_root(geobody_feature_root),
                                 geobody_feature_root.attrib['uuid'],
                                 content_type = 'obj_GeobodyFeature',
                                 root = gi)

      if add_as_part:
         self.model.add_part('obj_GeobodyInterpretation', self.uuid, gi)
         if add_relationships:
            self.model.create_reciprocal_relationship(gi, 'destinationObject', geobody_feature_root, 'sourceObject')

      return gi


class WellboreInterpretation(BaseResqpy):
   """Class for RESQML Wellbore Interpretation organizational objects.

   RESQML documentation:

      May refer to one of these:

      * **Wellbore**. A unique, oriented path from the bottom of a drilled borehole to the surface of the earth.
        The path must not overlap or cross itself.
      * **Borehole**. A hole excavated in the earth as a result of drilling or boring operations. The borehole
        may represent the hole of an entire wellbore (when no sidetracks are present), or a sidetrack extension.
        A borehole extends from an originating point (the surface location for the initial borehole or kickoff point
        for sidetracks) to a terminating (bottomhole) point.
      * **Sidetrack**. A borehole that originates in another borehole as opposed to originating at the surface.

   """

   resqml_type = 'WellboreInterpretation'
   valid_domains = ('depth', 'time', 'mixed')

   def __init__(self,
                parent_model,
                root_node = None,
                uuid = None,
                title = None,
                is_drilled = None,
                wellbore_feature = None,
                domain = 'depth',
                extra_metadata = None):
      """Initialises a wellbore interpretation organisational object."""

      # note: will create a paired WellboreFeature object when loading from xml

      self.is_drilled = is_drilled
      self.wellbore_feature = wellbore_feature
      self.feature_root = None if self.wellbore_feature is None else self.wellbore_feature.root
      if (not title) and self.wellbore_feature is not None:
         title = self.wellbore_feature.feature_name
      self.domain = domain
      super().__init__(model = parent_model,
                       uuid = uuid,
                       title = title,
                       extra_metadata = extra_metadata,
                       root_node = root_node)

   def _load_from_xml(self):
      root_node = self.root
      self.is_drilled = rqet.find_tag_bool(root_node, 'IsDrilled')
      self.domain = rqet.find_tag_text(root_node, 'Domain')
      interp_feature_ref_node = rqet.find_tag(root_node, 'InterpretedFeature')
      if interp_feature_ref_node is not None:
         self.feature_root = self.model.referenced_node(interp_feature_ref_node)
         if self.feature_root is not None:
            self.wellbore_feature = WellboreFeature(self.model,
                                                    uuid = self.feature_root.attrib['uuid'],
                                                    feature_name = self.model.title_for_root(self.feature_root))

   def iter_trajectories(self):
      """ Iterable of associated trajectories """

      import resqpy.well

      uuids = self.model.uuids(obj_type = "WellboreTrajectoryRepresentation", related_uuid = self.uuid)
      for uuid in uuids:
         yield resqpy.well.Trajectory(self.model, uuid = uuid)

   def is_equivalent(self, other, check_extra_metadata = True):
      """Returns True if this interpretation is essentially the same as the other; otherwise False."""

      if other is None or not isinstance(other, WellboreInterpretation):
         return False
      if self is other or bu.matching_uuids(self.uuid, other.uuid):
         return True
      if self.wellbore_feature is not None:
         if not self.wellbore_feature.is_equivalent(other.wellbore_feature):
            return False
      elif other.wellbore_feature is not None:
         return False
      if self.root is not None and other.root is not None:
         if rqet.citation_title_for_node(self.root) != rqet.citation_title_for_node(other.root):
            return False
         if self.domain != other.domain:
            return False
      elif self.root is not None or other.root is not None:
         return False
      if check_extra_metadata and not equivalent_extra_metadata(self, other):
         return False
      return (self.title == other.title and self.is_drilled == other.is_drilled)

   def create_xml(self,
                  wellbore_feature_root = None,
                  add_as_part = True,
                  add_relationships = True,
                  originator = None,
                  title_suffix = None,
                  reuse = True):
      """Creates a wellbore interpretation organisational xml node from a wellbore interpretation object."""

      # note: related wellbore feature node should be created first and referenced here

      if not self.title:
         self.title = self.wellbore_feature.feature_name
      if title_suffix:
         self.title += ' ' + title_suffix

      if reuse and self.try_reuse():
         return self.root
      wi = super().create_xml(add_as_part = False, originator = originator)

      if self.wellbore_feature is not None:
         wbf_root = self.wellbore_feature.root
         if wbf_root is not None:
            if wellbore_feature_root is None:
               wellbore_feature_root = wbf_root
            else:
               assert wbf_root is wellbore_feature_root, 'wellbore feature mismatch'

      if self.is_drilled is None:
         self.is_drilled = False

      id_node = rqet.SubElement(wi, ns['resqml2'] + 'IsDrilled')
      id_node.set(ns['xsi'] + 'type', ns['xsd'] + 'boolean')
      id_node.text = str(self.is_drilled).lower()

      assert self.domain in self.valid_domains, 'illegal domain value for wellbore interpretation'
      domain_node = rqet.SubElement(wi, ns['resqml2'] + 'Domain')
      domain_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Domain')
      domain_node.text = str(self.domain).lower()

      self.model.create_ref_node('InterpretedFeature',
                                 self.model.title_for_root(wellbore_feature_root),
                                 wellbore_feature_root.attrib['uuid'],
                                 content_type = 'obj_WellboreFeature',
                                 root = wi)

      if add_as_part:
         self.model.add_part('obj_WellboreInterpretation', self.uuid, wi)
         if add_relationships:
            self.model.create_reciprocal_relationship(wi, 'destinationObject', wellbore_feature_root, 'sourceObject')

      return wi



# function for reading a zmap file

version = '13th May 2021'

# RMS and ROXAR are registered trademarks of Roxar Software Solutions AS, an Emerson company

import logging

log = logging.getLogger(__name__)
log.debug('zmap_reader.py version ' + version)

import numpy as np


def read_zmap_header(inputfile):

   # read header, read until second '@', record header lines and content
   with open(inputfile, 'r') as infile:
      comments = []
      head = []
      a = 0  # count '@'s
      headers = 0  # count header+comment lines
      while a < 2:
         line = infile.readline()
         headers += 1
         if line[:1] == '@':
            a += 1
         if line[:1] == '!':
            comments.append(line.strip())
         elif a == 1 or (line[:1] == '@'):
            row = line.strip().strip('@').split(',')
            head.append(row)
         else:
            log.error("Header section does not seem to be defined by 2 @s")
            return None, None, None
      line = infile.readline()
      if line[0] == '+':
         headers = headers + 1  # add extra line for the + symbol..

   for c in comments:
      log.debug(c)

   # ok now process the header
#  nodes_per_line                          = int(head[0][2])
#  field_w                                 = head[1][0]
   null_value = head[1][1].strip()
   null_value2 = head[1][2].strip()
   #  dec_places                              = head[1][3]
   #  strt_c                                  = head[1][4]
   no_rows = int(head[2][0])
   no_cols = int(head[2][1])
   minx = np.float64(head[2][2])
   maxx = np.float64(head[2][3])
   miny = np.float64(head[2][4])
   maxy = np.float64(head[2][5])

   # decide on the null value
   if not null_value:
      null_value = null_value2
   log.debug("Read {} header lines, we have {} rows, {} cols, and data from {} to {} and {} to {}".format(
      headers, no_rows, no_cols, minx, maxx, miny, maxy))

   return headers, no_rows, no_cols, minx, maxx, miny, maxy, null_value


# note: the RMS text format was previously known as the Roxar format


def read_roxar_header(inputfile):

   with open(inputfile, 'r') as fp:
      words = fp.readline().split()
      assert words[0] == '-996', 'RMS text format indicator -996 not found'
      no_rows = int(words[1])
      # dx = float(words[2])
      # dy = float(words[3])
      words = fp.readline().split()
      minx = float(words[0])
      maxx = float(words[1])
      miny = float(words[2])
      maxy = float(words[3])
      words = fp.readline().split()
      no_cols = int(words[0])
   headers = 4
   null_value = '9999900.0000'
   return headers, no_rows, no_cols, minx, maxx, miny, maxy, null_value


def read_mesh(inputfile, dtype = np.float64, format = None):

   if format == 'zmap':
      headers, no_rows, no_cols, minx, maxx, miny, maxy, null_value = read_zmap_header(inputfile)
   elif format in ['rms', 'roxar']:
      headers, no_rows, no_cols, minx, maxx, miny, maxy, null_value = read_roxar_header(inputfile)
   else:
      raise ValueError('format not recognised for read_mesh: ' + str(format))
   # load the values in, converting null value to NaN's

   infile = open(inputfile, 'r')
   indata = infile.readlines()
   infile.close()
   c = 0
   f = np.zeros(no_cols * no_rows, dtype = dtype)
   n = 0
   for line in indata:
      c += 1
      if c <= headers:
         continue
      else:
         for x in line.split():
            if (x == null_value) or (dtype(x) == dtype(null_value)):
               f[n] = np.nan
            else:
               f[n] = dtype(x)
            n += 1
      if (n % 10000 == 0) and (n > 10000):
         log.debug("Read node {} of {}".format(n, no_cols * no_rows))

   log.debug("Read {} nodes".format(len(f)))

   if format == 'zmap':
      # reshape it, and swap axis. Beacuse columns major order from fortran.
      f = f.reshape((no_cols, no_rows)).swapaxes(0, 1)
   else:  # format in ['rms', 'roxar']
      f = f.reshape((no_rows, no_cols))

   # now generate x and y coords
   x = np.linspace(minx, maxx, no_cols)  # get x axis coords
   if format == 'zmap':
      y = np.linspace(maxy, miny, no_rows)
   else:  # format in ['rms', 'roxar']
      y = np.linspace(miny, maxy, no_rows)
   x, y = np.meshgrid(x, y)  # get x and y of every node
   assert x.shape == y.shape == f.shape

   return x, y, f


def read_zmapplusgrid(inputfile, dtype = np.float64):
   """Read zmapplus grid (surface mesh); returns triple (x, y, z) 2D arrays."""

   return read_mesh(inputfile, dtype = dtype, format = 'zmap')


def read_roxar_mesh(inputfile, dtype = np.float64):
   """Read RMS text format surface mesh; returns triple (x, y, z) 2D arrays.

   note:
      the RMS text format was previously known as the Roxar format
   """

   return read_mesh(inputfile, dtype = dtype, format = 'rms')


def read_rms_text_mesh(inputfile, dtype = np.float64):
   """Read RMS text format surface mesh; returns triple (x, y, z) 2D arrays.

   note:
      the RMS text format was previously known as the Roxar format
   """

   return read_roxar_mesh(inputfile, dtype = dtype)



"""volume.py: Functions to calculate volumes of hexahedral cells; assumes consistent length units."""

version = '31st August 2021'

import numpy as np


def _pyr(ra, ab, ac, ad, db):  # returns 6 * volume of one quad pyramid, defined by specific vectors
   return np.dot(ra, np.cross(db, ac)) + 0.5 * np.dot(ac, np.cross(ad, ab))


def pyramid_volume(apex, a, b, c, d, crs_is_right_handed = False):
   """Returns volume of a quadrilateral pyramid.

   arguments:
      apex (triple float): location of the apex of the pyramid
      a, b, c, d (each triple float): locations of corners of base of pyramid; clockwise viewed from apex
         for a left handed crs
      crs_is_right_handed (boolean, default False): set True if xyz axes of crs are right handed

   returns:
      float, being the volume of the pyramid; units are implied by crs units in use by the vertices
   """

   apex = np.array(apex)
   a = np.array(a)
   b = np.array(b)
   c = np.array(c)
   d = np.array(d)
   v = _pyr(a - apex, b - a, c - a, d - a, b - d) / 6.0
   if not crs_is_right_handed:
      v = -v
   return v


def tetrahedron_volume(a, b, c, d):
   """Returns volume of a tetrahedron.

   arguments:
      a, b, c, d (each triple float): locations of corners of tetrahedron
      crs_is_right_handed (boolean, default False): set True if xyz axes of crs are right handed

   returns:
      float, being the volume of the tetrahedron; units are implied by crs units in use by the vertices
   """

   a = np.array(a)
   b = np.array(b)
   c = np.array(c)
   d = np.array(d)

   return np.abs(np.dot(a - d, np.cross(b - d, c - d))) / 6.0


def tetra_cell_volume(cp, centre = None, off_hand = False):
   """Returns volume of single hexahedral cell with corner points cp of shape (2, 2, 2, 3); assumes bilinear faces."""

   if centre is None:
      centre = np.mean(cp.reshape((8, 3)), axis = 0)
   r0 = cp[0, 0, 0] - centre
   r1 = cp[1, 1, 1] - centre
   v = 0.0
   v += _pyr(r0, cp[0, 0, 1] - cp[0, 0, 0], cp[0, 1, 1] - cp[0, 0, 0], cp[0, 1, 0] - cp[0, 0, 0],
             cp[0, 1, 0] - cp[0, 0, 1])
   v += _pyr(r0, cp[1, 0, 0] - cp[0, 0, 0], cp[1, 0, 1] - cp[0, 0, 0], cp[0, 0, 1] - cp[0, 0, 0],
             cp[0, 0, 1] - cp[1, 0, 0])
   v += _pyr(r0, cp[0, 1, 0] - cp[0, 0, 0], cp[1, 1, 0] - cp[0, 0, 0], cp[1, 0, 0] - cp[0, 0, 0],
             cp[1, 0, 0] - cp[0, 1, 0])
   v += _pyr(r1, cp[1, 0, 1] - cp[1, 1, 1], cp[1, 0, 0] - cp[1, 1, 1], cp[1, 1, 0] - cp[1, 1, 1],
             cp[1, 1, 0] - cp[1, 0, 1])
   v += _pyr(r1, cp[1, 1, 0] - cp[1, 1, 1], cp[0, 1, 0] - cp[1, 1, 1], cp[0, 1, 1] - cp[1, 1, 1],
             cp[0, 1, 1] - cp[1, 1, 0])
   v += _pyr(r1, cp[0, 1, 1] - cp[1, 1, 1], cp[0, 0, 1] - cp[1, 1, 1], cp[1, 0, 1] - cp[1, 1, 1],
             cp[1, 0, 1] - cp[0, 1, 1])

   if off_hand:
      v = -v

   return v / 6.0


def tetra_volumes_slow(cp, centres = None, off_hand = False):
   """Returns volume array for all hexahedral cells assuming bilinear faces, using loop over cells."""

   # NB: deprecated, superceded by much faster function below
   # todo: handle NaNs
   # Pagoda style corner point data
   assert cp.ndim == 7

   flat = cp.reshape(-1, 2, 2, 2, 3)
   cells = flat.shape[0]
   if centres is None:
      centres = np.mean(flat, axis = (1, 2, 3))
   else:
      centres = centres.reshape((-1, 3))
   volumes = np.zeros(cells)
   for cell in range(cells):
      volumes[cell] = tetra_cell_volume(flat[cell], centre = centres[cell], off_hand = off_hand)
   return volumes.reshape(cp.shape[0:3])


def tetra_volumes(cp, centres = None, off_hand = False):
   """Returns volume array for all hexahedral cells assuming bilinear faces, using numpy operations.

   arguments:
      cp (7D numpy array of floats): cell corner point data in Pagoda 7D format [nk, nj, ni, kp, jp, ip, xyz]
      centres (optional, 4D numpy array of floats): cell centre points [nk, nj, ni, xyz]; calculated if None
      off_hand (boolean, default False): if True, the handedness of IJK space is the opposite of that for xyz
         space; if this argument is not set correctly, negative volumes will be returned

   returns:
      numpy 3D array of floats being the cell volumes [nk, nj, ni]

   note:
      length units are assumed to be consistent in x, y & z; and untis of returned volumes are implicitly
      those length units cubed
   """

   def pyrs(ra, ab, ac, ad, db):  # returns 6 * volume of one quad pyramid (for each cell)
      return np.sum(ra * np.cross(db, ac), axis = -1) + 0.5 * np.sum(ac * np.cross(ad, ab), axis = -1)

   # Pagoda style corner point data
   assert cp.ndim == 7

   flat = cp.reshape(-1, 2, 2, 2, 3)
   cells = flat.shape[0]
   if centres is None:
      centres = np.mean(flat, axis = (1, 2, 3))
   else:
      centres = centres.reshape((-1, 3))
   v = np.zeros(cells)
   # todo: numpy array operation covering all cells in grid

   r0 = flat[:, 0, 0, 0] - centres
   r1 = flat[:, 1, 1, 1] - centres

   v += pyrs(r0, flat[:, 0, 0, 1] - flat[:, 0, 0, 0], flat[:, 0, 1, 1] - flat[:, 0, 0, 0],
             flat[:, 0, 1, 0] - flat[:, 0, 0, 0], flat[:, 0, 1, 0] - flat[:, 0, 0, 1])
   v += pyrs(r0, flat[:, 1, 0, 0] - flat[:, 0, 0, 0], flat[:, 1, 0, 1] - flat[:, 0, 0, 0],
             flat[:, 0, 0, 1] - flat[:, 0, 0, 0], flat[:, 0, 0, 1] - flat[:, 1, 0, 0])
   v += pyrs(r0, flat[:, 0, 1, 0] - flat[:, 0, 0, 0], flat[:, 1, 1, 0] - flat[:, 0, 0, 0],
             flat[:, 1, 0, 0] - flat[:, 0, 0, 0], flat[:, 1, 0, 0] - flat[:, 0, 1, 0])
   v += pyrs(r1, flat[:, 1, 0, 1] - flat[:, 1, 1, 1], flat[:, 1, 0, 0] - flat[:, 1, 1, 1],
             flat[:, 1, 1, 0] - flat[:, 1, 1, 1], flat[:, 1, 1, 0] - flat[:, 1, 0, 1])
   v += pyrs(r1, flat[:, 1, 1, 0] - flat[:, 1, 1, 1], flat[:, 0, 1, 0] - flat[:, 1, 1, 1],
             flat[:, 0, 1, 1] - flat[:, 1, 1, 1], flat[:, 0, 1, 1] - flat[:, 1, 1, 0])
   v += pyrs(r1, flat[:, 0, 1, 1] - flat[:, 1, 1, 1], flat[:, 0, 0, 1] - flat[:, 1, 1, 1],
             flat[:, 1, 0, 1] - flat[:, 1, 1, 1], flat[:, 1, 0, 1] - flat[:, 0, 1, 1])

   v /= 6.0

   if off_hand:
      v = np.negative(v)
   return v.reshape(cp.shape[0:3])
